#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DATASTART 0x2C

/**************/
/* Data types */
/**************/

typedef unsigned short int word;   /* %u */
typedef signed short int sword;   /* %u */
typedef unsigned long dword;       /* %lu */

/********************/
/* WAVE RIFF Header */
/********************/

/* Last four bytes contains (long int) length of PCM data chunk which
 * is unknown at start of recording. Offset is 0x28, header lenght is 0x2C
 */

/* http://ccrma.stanford.edu/CCRMA/Courses/422/projects/WaveFormat/ */
/* http://web.archive.org/web/20030807142512/sox.sourceforge.net/AudioFormats-11.html#ss11.6 */

typedef struct WAVEheader_contents
{
        char  ckID[4];             /* chunk id 'RIFF'            */
        dword ckSize;              /* chunk size                 */
        char  wave_ckID[4];        /* wave chunk id 'WAVE'       */
        char  fmt_ckID[4];         /* format chunk id 'fmt '     */
        dword fmt_ckSize;          /* format chunk size          */
        word  formatTag;           /* format tag currently pcm   */
        word  nChannels;           /* number of channels         */
        dword nSamplesPerSec;      /* sample rate in hz          */
        dword nAvgBytesPerSec;     /* average bytes per second   */
        word  nBlockAlign;         /* number of bytes per sample */
        word  nBitsPerSample;      /* number of bits in a sample */
        char  data_ckID[4];        /* data chunk id 'data'       */
        dword data_ckSize;         /* length of data chunk       */
} WAVEheader;

typedef struct Audiodata_contents {
		unsigned int counter;
		float time;
		float value;
} Audiodata;


void PrintWAVinfo(WAVEheader *WAV_file){
	printf("# Format Chunk Size: %lu\n", WAV_file->fmt_ckSize);
	printf("# Format Type: ");
		if(WAV_file->formatTag == 1)	printf("PCM (0x01)\n");
		else printf("unknown compression (0x%x)\n", WAV_file->formatTag);
	printf("# Number of Channels: %u ",  WAV_file->nChannels);
		if(WAV_file->nChannels == 1)	printf("(mono)\n");
		if(WAV_file->nChannels == 2)	printf("(stereo)\n");

	printf("# Sample rate: %lu Hz\n",  WAV_file->nSamplesPerSec);
	printf("# Average bytes per second: %lu\n", WAV_file->nAvgBytesPerSec);
	printf("# Number of Bytes per Sample: %u\n", WAV_file->nBlockAlign);
	printf("# Number of Bits per Sample: %u\n", WAV_file->nBitsPerSample);
	printf("# Lenght of Data Chunk: %lu\n", WAV_file->data_ckSize );
}

void *ReadData(FILE *file, word data_size, dword chunk_size){
    word *data;

/*	printf("Data size: %u\n", data_size);
	printf("Size of block: %lu\n", chunk_size);
*/

	data = (void *) malloc(chunk_size * data_size);
	if(!data) return NULL;

	fread(data, chunk_size, 1, file);  

	return data;
}

void Amplitude(void *stream, dword chunk_size, word data_size, dword sample_rate){
	dword i = 0, nsample;
	/* PCM data */ 
	unsigned char A11 = 0;
	sword A21 = 0;
	float Time, Amplitude;
	
	/* Number of samples for reading */
	nsample = chunk_size / data_size;
	
	while(i != nsample){
		Time = (float) i / (float) sample_rate;
		switch(data_size)
		{
			case 1: /* 8-bits  -- unsigned char */
				A11 = *(unsigned char *) ((unsigned char *) stream + i);
				Amplitude = (float) (A11 - 0x7F);
				break;
			case 2: /* 16-bits */
				/* 16-bit samples are stored as 2's-complement 
				   signed integers, ranging from -32768 to 32767 */
				A21 = *(sword *) ((sword *) stream + i);
				Amplitude = (float) (A21);  
				break;
		}	
		printf("%f %f\n", Time, Amplitude); 
		i = i + 1;
	} 

}

void Sine2Square(void *stream, dword chunk_size, word data_size){
	dword i = 0, nsample;
	sword A1 = 0;
	
	nsample = chunk_size / data_size;
	
	while(i != nsample){
		switch(data_size)
		{
			case 1: /* 8-bits */
				A1 = *(char *) ((char *) stream + i);

				if(A1 > 0x7F)
					*((char *) stream + i) = (char) 0x00;
				else if (A1 < 0x7F) 
					*((char *) stream + i) = (char) 0xFF;
				else 
					*((char *) stream + i) = (char) 0x7F;

				break;
			case 2: /* 16-bits */
				A1 = *(sword *) ((sword *) stream + i);
				if( A1 > 0 )
					*((sword *) stream + i) = (sword) 32767;
				else if ( A1 < 0 )
					*((sword *) stream + i) = (sword) -32767;
				else
					*((sword *) stream + i) = (sword) 0;

				break;
		}	
		i = i + 1;
	}
}

float Differenciator(sword x1, sword x2, sword y1, sword y2){
	float gradient = ((float) (y2 - y1)) / ((float) (x2 - x1));
	return (-y1 / gradient);
}

unsigned int 
ZeroDetector(Audiodata *sample_data, void *stream, dword chunk_size, word data_size, dword sample_rate){
/* Returns number of collected sample_data */
	unsigned int counter = 1;
	dword nsample, i = 0;
	sword A1 = 0, A2 = 0; /* PCM data */ 
	float z1 = 0.0, z2 = 0.0; /* Zero-cross */
	char zcross = 0;
	float Time, Frequency, Diff;
	
	/* Number of samples for reading */
	nsample = chunk_size / data_size;
	
	while(i != nsample){
		switch(data_size)
		{
			case 1: /* 8-bits */
				A1 = *(unsigned char *) ((unsigned char *) stream + i);
				A2 = *(unsigned char *) ((unsigned char *) stream + i+1);
				if( ((A1 < 0x7F) && (A2 > 0x7F )) || ((A1 > 0x7F) && (A2 < 0x7F )))
				{
					Diff = Differenciator((sword) i, (sword) i+1, A1, A2);
					z1 = z2;
					z2 = i + Diff;
					zcross = 1;
				}
				break;
			case 2: /* 16-bits */
				A1 = *(sword *) ((sword *) stream + i);
				A2 = *(sword *) ((sword *) stream + i+1); 

				if( ((A1 <= 0) && (A2 > 0 )) || ((A1 > 0) && (A2 <= 0 )))
				{
					Diff = Differenciator((sword) i, (sword) i+1, A1, A2);
					z1 = z2;
					z2 = i + Diff;
					zcross = 1;
				}
				break;

		}	
		/* Count Frequency */
		if(zcross == 1){
			Frequency = (float)  sample_rate / (2.0 * (z2 - z1) ); 
			Time = (float) i / (float) sample_rate;

/*			printf("%lu %lu %lu %lu %f \t %f\n", i, z2, z1, z2 - z1, Time, Frequency);*/
/*			printf("%f %f\n", Time, Frequency); */
			sample_data->counter = counter;
			sample_data->time = Time;
			sample_data->value = Frequency;

			sample_data++; counter++; zcross = 0;
		}
		i = i + 1;
	} 
	return counter;
}

void PrintOut(Audiodata *sample_data, unsigned int counter){
	unsigned int i;

	printf("#----------------------------------------------------------------\n");
	printf("#Number\tTime\tValue\n");
	printf("#----------------------------------------------------------------\n");
	for(i = 0; i <= counter; i++){
		if(sample_data->counter != 0)
			printf("%d %f %f\n", sample_data->counter, sample_data->time, sample_data->value);
		sample_data++;
	}
}

void SSTV_Filter(Audiodata *sample_data, unsigned int counter){
	unsigned int i;
	for(i = 0; i <= counter; i++){
		if(sample_data->value < 1090.0 || sample_data->value > 2350.0 )
			sample_data->counter = 0;
		sample_data++;
	}

}

/* ***********************
 *                       *
 *                       *
 *        M A I N        *
 *                       *
 *                       *
 *********************** */

int main(int argc, char *argv[]){
	FILE *fp;
/*	int options; */
	unsigned int counter;
	char *FILENAME;
	WAVEheader header;
	dword DATAS, SSIZE, SRATE;
	word *sound_data;
	Audiodata *sample_data;	
	
	if(argc < 1){
		return 0;
	}

	FILENAME = argv[1];
	
	/* Reset header */
	memset(&header, 0, sizeof(header));
	
	fp = fopen(FILENAME, "rb");
	if(!fp){
		fprintf(stderr, "File '%s' cannot be opened!\n", FILENAME);
		return -1;
	}		

	/* Read Header */ /*fseek(fp, 0, SEEK_SET);  */
	fread(&header, sizeof(header), 1, fp);
	
	if(strncmp(header.ckID, "RIFF", 4) != 0){
		fprintf(stderr, "This file is not RIFF WAVe format!\n");
		fclose(fp);
		return(-1);
	};

	printf("# Analyzed file: %s\n", FILENAME);
	PrintWAVinfo(&header); 

	/* Size of all data */ 
	DATAS = header.data_ckSize;
	/* Size of single word */
	SSIZE = header.nBlockAlign;
	/* Sample rate */
	SRATE = header.nSamplesPerSec;

	sound_data = ReadData(fp, SSIZE, DATAS);

	if(!sound_data){
		fprintf(stderr, "Memory allocation error!\n");
		fclose(fp);
		return(-1);
	}

#if 0
	while(options = getopt("fai:w:o:") != -1 ){
		switch(options){
			case 'i': /* Print file info */
				break;
			case 'w': /* Input file */
				break;
			case 'o': /* Output file */
				break;
			case 'a': /* Amplitude data */
				break;
			case 'f': /* Frequncy data */
				break;
		}
	}
#endif
	
/*	Sine2Square(sound_data, DATAS, SSIZE);  */

	if( ( sample_data = (Audiodata *) malloc(DATAS*sizeof(Audiodata)) ) == NULL  )
	{
		fprintf(stderr, "Out of memory!\n");
		return -1;
	}

	counter = ZeroDetector(sample_data, sound_data, DATAS, SSIZE, SRATE); 
	fprintf(stderr, "Collected: %d values.\n", counter);

	SSTV_Filter(sample_data, counter); 
	PrintOut(sample_data, counter);
/*	Amplitude(sound_data, DATAS, SSIZE, SRATE); */

	free(sample_data);
	free(sound_data);
	fclose(fp);
	return 0;
}


